from django.db import IntegrityError
from django.test import TestCase

from wanderer.models import WandererManagedMap
from wanderer.tests.utils import create_managed_map, create_wanderer_users


class TestMap(TestCase):

    def test_same_map(self):
        """Cancels creating 2 maps with the same name/slug combination"""
        WandererManagedMap.objects.create(
            wanderer_url="fake_url",
            map_slug="slug",
            map_api_key="fake_key",
            map_acl_id="id",
            map_acl_api_key="fake_key",
        )

        self.assertRaises(
            IntegrityError,
            WandererManagedMap.objects.create,
            wanderer_url="fake_url",
            map_slug="slug",
            map_api_key="fake_key",
            map_acl_id="id",
            map_acl_api_key="fake_key",
        )

    def test_get_all_accounts_characters_ids(self):
        managed_map = create_managed_map()
        create_wanderer_users(managed_map, 2)

        all_character_ids = managed_map.get_all_accounts_characters_ids()

        self.assertEqual(len(all_character_ids), 4)
        self.assertIn(1000, all_character_ids)
        self.assertIn(1001, all_character_ids)
        self.assertIn(1010, all_character_ids)
        self.assertIn(1011, all_character_ids)



"""Models."""

from typing import Optional

from django.conf import settings
from django.contrib.auth.models import Group, User
from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import State
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
    EveFactionInfo,
)
from allianceauth.framework.api.user import get_all_characters_from_user
from allianceauth.services.hooks import get_extension_logger

from wanderer.managers import WandererManagedMapManager
from wanderer.wanderer import (
    NotFoundError,
    add_character_to_acl,
    get_acl_members,
    remove_member_from_access_list,
)

logger = get_extension_logger(__name__)


class General(models.Model):
    """A metamodel for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (("basic_access", "Can access this app"),)


class WandererManagedMap(models.Model):
    """Wanderer map with an ACL managed by the auth"""

    objects = WandererManagedMapManager()

    name = models.CharField(
        max_length=80,
        help_text=_("User friendly name for your users to recognize the map"),
    )
    wanderer_url = models.CharField(
        max_length=120, help_text=_("URL of the wanderer instance")
    )
    map_slug = models.CharField(
        max_length=20, help_text=_("Map slug on the wanderer instance")
    )
    map_api_key = models.CharField(max_length=100, help_text=_("API key of the map"))

    map_acl_id = models.CharField(
        max_length=100, help_text=_("ID of the managed access list")
    )
    map_acl_api_key = models.CharField(
        max_length=100, help_text=_("API key of the managed access list")
    )

    state_access = models.ManyToManyField(
        State, blank=True, help_text=_("States to whose members this map is available.")
    )

    group_access = models.ManyToManyField(
        Group, blank=True, help_text=_("Groups to whose members this map is available.")
    )

    character_access = models.ManyToManyField(
        EveCharacter,
        blank=True,
        help_text=_("Characters to which this map is available."),
    )

    corporation_access = models.ManyToManyField(
        EveCorporationInfo,
        blank=True,
        help_text=_("Corporations to whose members this map is available."),
    )

    alliance_access = models.ManyToManyField(
        EveAllianceInfo,
        blank=True,
        help_text=_("Alliances to whose members this map is available."),
    )

    faction_access = models.ManyToManyField(
        EveFactionInfo,
        blank=True,
        help_text=_("Factions to whose members this map is available."),
    )

    def __str__(self):
        return f"{self.wanderer_url}/{self.map_slug}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["wanderer_url", "map_slug"], name="functional_pk_urlslug"
            )
        ]

    def accessible_by(self, user: User) -> bool:
        """Defines if a user can access this map or not"""

        logger.debug("Checking if user %s can access the map %s", user, self.name)

        if not user.has_perm("wanderer.basic_access"):
            return False

        try:
            main_character: EveCharacter = user.profile.main_character
            assert main_character

            if user.is_superuser:
                logger.info("Returning all servers to user %s", user)
                return True

            # build queries then OR them all
            queries = []

            # States access everyone has a state
            queries.append(models.Q(state_access=user.profile.state))
            # Groups access, is ok if no groups.
            queries.append(models.Q(group_access__in=user.groups.all()))
            # ONLY on main char from here down
            # Character access
            queries.append(models.Q(character_access=main_character))
            # Corp access
            try:
                queries.append(
                    models.Q(
                        corporation_access=EveCorporationInfo.objects.get(
                            corporation_id=main_character.corporation_id
                        )
                    )
                )
            except EveCorporationInfo.DoesNotExist:
                pass
            # Alliance access if part of an alliance
            try:
                if main_character.alliance_id:
                    queries.append(
                        models.Q(
                            alliance_access=EveAllianceInfo.objects.get(
                                alliance_id=main_character.alliance_id
                            )
                        )
                    )
            except EveAllianceInfo.DoesNotExist:
                pass
            # Faction access if part of a faction
            try:
                if main_character.faction_id:
                    queries.append(
                        models.Q(
                            faction_access=EveFactionInfo.objects.get(
                                faction_id=main_character.faction_id
                            )
                        )
                    )
            except EveFactionInfo.DoesNotExist:
                pass

            logger.debug(
                "%d queries for %s 's visible characters", len(queries), main_character
            )

            if settings.DEBUG:
                logger.debug(queries)

            # filter based on "OR" all queries
            query = queries.pop()
            for q in queries:
                query |= q
            return WandererManagedMap.objects.filter(query, id=self.id).exists()

        except AssertionError:
            logger.info("User %s without eve character can't access maps", user)
            return False

    def user_has_account(self, user: User) -> bool:
        """Return true if the user has an active account on this map"""
        return WandererAccount.objects.filter(user=user, wanderer_map=self).exists()

    def get_user_account(self, user: User) -> Optional["WandererAccount"]:
        """Returns the user account associated to this map if it exists"""
        try:
            return WandererAccount.objects.get(user=user, wanderer_map=self)
        except WandererAccount.DoesNotExist:
            return None

    def delete_user(self, user: User):
        """Removes the user characters from the map and then deletes the associated account"""
        wanderer_account = self.get_user_account(user)
        for character_to_remove_id in wanderer_account.get_all_character_ids():
            try:
                self.remove_member_from_access_list(character_to_remove_id)
            except (
                NotFoundError
            ):  # If the character is already off the access list we're good
                pass
        wanderer_account.delete()

    def get_character_ids_on_access_list(self) -> list[int]:
        """Returns all character_ids present on the access list"""
        return get_acl_members(self.wanderer_url, self.map_acl_id, self.map_acl_api_key)

    def add_character_to_acl(self, character_id: int):
        """Adds a single character to the ACL with the viewer role"""
        return add_character_to_acl(
            self.wanderer_url, self.map_acl_id, self.map_acl_api_key, character_id
        )

    def remove_member_from_access_list(self, member_id: int):
        """
        Removes a member from the access list.
        member_id can be character/corporation/alliance.
        """

        return remove_member_from_access_list(
            self.wanderer_url, self.map_acl_id, self.map_acl_api_key, member_id
        )

    def get_all_accounts_characters_ids(self) -> list[int]:
        """
        Returns a list of all character ids of accounts linked to this map
        """
        return list(
            self.accounts.values_list(
                "user__character_ownerships__character__character_id", flat=True
            )
        )


class WandererAccount(models.Model):
    """Represents a user linked to a wanderer map"""

    user = models.ForeignKey(
        User, on_delete=models.CASCADE, help_text=_("Auth user linked to the map")
    )
    wanderer_map = models.ForeignKey(
        WandererManagedMap,
        models.CASCADE,
        related_name="accounts",
        related_query_name="account",
        help_text=_("Wanderer map to which the user is linked"),
    )

    def __str__(self):
        return f"{self.user} - {self.wanderer_map}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "wanderer_map"], name="functional_pk_user_map"
            )
        ]

    def get_all_character_ids(self) -> list[int]:
        """Return all character ids associated to this account"""
        return [
            character.character_id
            for character in get_all_characters_from_user(self.user)
        ]



"""Tasks tests"""

from unittest.mock import MagicMock, call

from django.test import TestCase

from wanderer.models import WandererAccount, WandererManagedMap
from wanderer.tasks import (
    add_alts_to_map,
    cleanup_access_list,
    remove_user_characters_from_map,
)

from .utils import create_managed_map, create_wanderer_users


class TestTasks(TestCase):

    def test_add_character_to_acl(self):
        """Checks that characters properly get added to the access list"""
        WandererManagedMap.get_character_ids_on_access_list = MagicMock(
            return_value=[1001, 1002]
        )  # Creating fake ids to be returned
        WandererManagedMap.add_character_to_acl = MagicMock()
        WandererAccount.get_all_character_ids = MagicMock(
            return_value=[1001, 1003]
        )  # Missing id 1003

        wanderer_map = create_managed_map()
        user = create_wanderer_users(wanderer_map)[0]

        add_alts_to_map(user.id, wanderer_map.id)

        WandererManagedMap.get_character_ids_on_access_list.assert_called_once()
        WandererManagedMap.add_character_to_acl.assert_called_once_with(1003)
        WandererAccount.get_all_character_ids.assert_called_once()

    def test_remove_user_characters_from_acl(self):
        WandererManagedMap.get_character_ids_on_access_list = MagicMock(
            return_value=[1001, 1002, 1003]
        )
        WandererManagedMap.remove_member_from_access_list = MagicMock()
        WandererAccount.get_all_character_ids = MagicMock(return_value=[1001, 1002])

        wanderer_map = create_managed_map()
        user = create_wanderer_users(wanderer_map)[0]

        remove_user_characters_from_map(user.id, wanderer_map.id)

        WandererManagedMap.get_character_ids_on_access_list.assert_called_once()
        WandererAccount.get_all_character_ids.assert_called_once()

        remove_member_calls = [call(1001), call(1002)]
        WandererManagedMap.remove_member_from_access_list.assert_has_calls(
            remove_member_calls, any_order=True
        )

    def test_cleanup_access_list(self):
        WandererManagedMap.get_character_ids_on_access_list = MagicMock(
            return_value=[1000, 1011, 1020]
        )
        WandererManagedMap.remove_member_from_access_list = MagicMock()
        WandererManagedMap.add_character_to_acl = MagicMock()

        wanderer_map = create_managed_map()
        create_wanderer_users(wanderer_map, 2)

        cleanup_access_list(wanderer_map.id)

        WandererManagedMap.get_character_ids_on_access_list.assert_called_once()
        WandererManagedMap.remove_member_from_access_list.assert_called_once_with(1020)
        add_character_calls = [call(1001), call(1010)]
        WandererManagedMap.add_character_to_acl.assert_has_calls(
            add_character_calls, any_order=True
        )



"""Admin site."""

from django.contrib import admin

from wanderer.models import WandererAccount, WandererManagedMap
from wanderer.wanderer import create_acl_associated_to_map


@admin.register(WandererManagedMap)
class WandererManagedMapAdmin(admin.ModelAdmin):
    readonly_fields = ["map_acl_id", "map_acl_api_key"]
    fieldsets = [
        (None, {"fields": ["name", "wanderer_url", "map_slug", "map_api_key"]}),
        ("Access List", {"fields": ["map_acl_id", "map_acl_api_key"]}),
        (
            "Access",
            {
                "fields": [
                    "state_access",
                    "group_access",
                    "character_access",
                    "corporation_access",
                    "alliance_access",
                ]
            },
        ),
    ]

    def save_model(self, request, obj, form, change):
        if not change:  # Only on item creation
            character = request.user.profile.main_character
            w = form.save(commit=False)
            map_acl_id, map_acl_api_key = create_acl_associated_to_map(
                w.wanderer_url, w.map_slug, character.character_id, w.map_api_key
            )
            w.map_acl_id = map_acl_id
            w.map_acl_api_key = map_acl_api_key
            w.save()
        else:
            super().save_model(request, obj, form, change)


@admin.register(WandererAccount)
class WandererUserAdmin(admin.ModelAdmin):
    list_filter = ["wanderer_map"]
    list_display = ["user", "wanderer_map"]
    readonly_fields = ["user", "wanderer_map"]

    def has_add_permission(self, request):
        return False



from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_fake_user

from wanderer.models import WandererAccount, WandererManagedMap


def create_managed_map() -> WandererManagedMap:
    """Create a basic for testing"""
    return WandererManagedMap.objects.create(
        wanderer_url="http://wanderer.localhost",
        map_slug="test",
        map_api_key="bad-map-api-key",
        map_acl_id="ACL_UUID",
        map_acl_api_key="bad-acl-api-key",
    )


def create_wanderer_users(
    wanderer_managed_map: WandererManagedMap, count=1
) -> list[WandererAccount]:
    """
    Create fake users linked to a map and character ownerships associated
    """
    users = []
    for i in range(count):
        user = create_fake_user(
            character_name=f"fake character {i}",
            character_id=1000 + i,
            corporation_id=2000 + i,
            corporation_name=f"fake corporation {i}",
            corporation_ticker=f"FAKE{i}",
        )
        users.append(
            WandererAccount.objects.create(user=user, wanderer_map=wanderer_managed_map)
        )

        for j in range(2):
            eve_character, _ = EveCharacter.objects.get_or_create(
                character_id=1000 + 10 * j + i,
                character_name=f"fake character {10*j+i}",
                corporation_id=2000 + i,
                corporation_name=f"fake corporation {i}",
                corporation_ticker=f"FAKE{i}",
            )
            char_ownership, _ = CharacterOwnership.objects.get_or_create(
                character=eve_character, user=user, owner_hash=f"fake hash {10*j+i}"
            )

    return users



# Generated by Django 4.2.16 on 2025-03-04 17:57

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("authentication", "0024_alter_userprofile_language"),
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (("basic_access", "Can access this app"),),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="WandererManagedMap",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="User friendly name for your users to recognize the map",
                        max_length=80,
                    ),
                ),
                (
                    "wanderer_url",
                    models.CharField(
                        help_text="URL of the wanderer instance", max_length=120
                    ),
                ),
                (
                    "map_slug",
                    models.CharField(
                        help_text="Map slug on the wanderer instance", max_length=20
                    ),
                ),
                (
                    "map_api_key",
                    models.CharField(help_text="API key of the map", max_length=100),
                ),
                (
                    "map_acl_id",
                    models.CharField(
                        help_text="ID of the managed access list", max_length=100
                    ),
                ),
                (
                    "map_acl_api_key",
                    models.CharField(
                        help_text="API key of the managed access list", max_length=100
                    ),
                ),
                (
                    "alliance_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Alliances to whose members this map is available.",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "character_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Characters to which this map is available.",
                        to="eveonline.evecharacter",
                    ),
                ),
                (
                    "corporation_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Corporations to whose members this map is available.",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "faction_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Factions to whose members this map is available.",
                        to="eveonline.evefactioninfo",
                    ),
                ),
                (
                    "group_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Groups to whose members this map is available.",
                        to="auth.group",
                    ),
                ),
                (
                    "state_access",
                    models.ManyToManyField(
                        blank=True,
                        help_text="States to whose members this map is available.",
                        to="authentication.state",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="WandererAccount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="Auth user linked to the map",
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "wanderer_map",
                    models.ForeignKey(
                        help_text="Wanderer map to which the user is linked",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="accounts",
                        related_query_name="account",
                        to="wanderer.wanderermanagedmap",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="wanderermanagedmap",
            constraint=models.UniqueConstraint(
                fields=("wanderer_url", "map_slug"), name="functional_pk_urlslug"
            ),
        ),
        migrations.AddConstraint(
            model_name="wandereraccount",
            constraint=models.UniqueConstraint(
                fields=("user", "wanderer_map"), name="functional_pk_user_map"
            ),
        ),
    ]



from django.core.management import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from wanderer.tasks import cleanup_all_access_lists

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    """Cleanup utility"""

    help = "Runs the cleanup command on all registered access lists"

    def handle(self, *args, **options):
        logger.info("Management command to cleanup access lists")
        cleanup_all_access_lists.delay()
        logger.info("Command success started")



"""Views."""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as _

from allianceauth.services.hooks import get_extension_logger

from wanderer.models import WandererAccount, WandererManagedMap
from wanderer.tasks import add_alts_to_map

logger = get_extension_logger(__name__)


@login_required
@permission_required("wanderer.basic_access")
def link(request, map_id: int):
    """Link a new user to a wanderer map"""
    wanderer_map = get_object_or_404(WandererManagedMap, pk=map_id)
    user = request.user

    if not wanderer_map.accessible_by(user):
        messages.warning(request, _("You don't have the access for this map"))
        logger.warning(
            "User id %d tried to access map id %d without authorization",
            user.id,
            wanderer_map.id,
        )

    elif wanderer_map.user_has_account(user):
        messages.warning(request, _("You are already linked to this map"))
    else:
        wanderer_user = WandererAccount.objects.create(
            user=user, wanderer_map=wanderer_map
        )
        add_alts_to_map.delay(wanderer_user.id, wanderer_map.id)
        messages.success(
            request,
            _(
                "Successfully linked your account to this map. Character update starting now."
            ),
        )

    return redirect("services:services")


@login_required
@permission_required("wanderer.basic_access")
def sync(request, map_id: int):
    """Checks that all the user characters are properly added to the access list"""
    wanderer_map = get_object_or_404(WandererManagedMap, pk=map_id)
    wanderer_user = get_object_or_404(
        WandererAccount, user=request.user, wanderer_map=wanderer_map
    )
    add_alts_to_map.delay(wanderer_user.id, wanderer_map.id)
    messages.success(request, _("Updating your characters with the map."))

    return redirect("services:services")


@login_required
@permission_required("wanderer.basic_access")
def remove(request, map_id: int):
    """Removes all characters from the map access list and deletes the user"""
    wanderer_map = get_object_or_404(WandererManagedMap, pk=map_id)
    user = request.user

    if not wanderer_map.user_has_account(user):
        messages.warning(request, _("You don't seem to be linked to this map."))
    else:
        wanderer_map.delete_user(user)

        messages.success(
            request, _("Successfully removed you from the map %s") % wanderer_map
        )

    return redirect("services:services")



"""Routes."""

from django.urls import include, path

from . import views

app_name = "wanderer"

module_urls = [
    path("link/<int:map_id>", views.link, name="link"),
    path("sync/<int:map_id>", views.sync, name="sync"),
    path("remove/<int:map_id>", views.remove, name="remove"),
]

urlpatterns = [path("wanderer/", include((module_urls, app_name), namespace=app_name))]



import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class WandererServiceConfig(AppConfig):
    name = "wanderer"
    label = "wanderer"
    verbose_name = "wanderer"

    def ready(self):
        """Runs service sync on startup"""
        # needs to be imported after the app is loaded
        from .auth_hooks import add_del_callback

        try:
            add_del_callback()
        except Exception as e:
            logger.error("WMM failed to initiate hooks")
            logger.error(e)



"""Tasks."""

from celery import chain, shared_task

from allianceauth.services.hooks import get_extension_logger

from wanderer.models import WandererAccount, WandererManagedMap

logger = get_extension_logger(__name__)


@shared_task
def add_alts_to_map(wanderer_user_id: int, wanderer_managed_map_id: int):
    """
    Ensures that all alts of this user are properly added on the wanderer map ACL.
    """
    logger.info(
        "Updating character of user id %d on map id %d",
        wanderer_user_id,
        wanderer_managed_map_id,
    )

    wanderer_user = WandererAccount.objects.get(pk=wanderer_user_id)
    wanderer_managed_map = WandererManagedMap.objects.get(pk=wanderer_managed_map_id)

    logger.debug("Recovered user %s and map %s", wanderer_user, wanderer_managed_map)

    characters_on_acl_ids_set = set(
        wanderer_managed_map.get_character_ids_on_access_list()
    )
    logger.debug(characters_on_acl_ids_set)
    user_character_ids_set = set(wanderer_user.get_all_character_ids())
    logger.debug(user_character_ids_set)

    missing_characters_set = user_character_ids_set - characters_on_acl_ids_set
    logger.info(
        "Need to add %d characters to the access list", len(missing_characters_set)
    )
    logger.debug(missing_characters_set)

    for missing_character_id in missing_characters_set:
        logger.debug("Adding character id %d to the access list", missing_character_id)
        wanderer_managed_map.add_character_to_acl(missing_character_id)


@shared_task
def remove_user_characters_from_map(
    wanderer_user_id: int, wanderer_managed_map_id: int
):
    """
    Removes all characters from that specific user from the map
    """
    logger.info(
        "Removing all characters of user id %d from map id %d",
        wanderer_user_id,
        wanderer_managed_map_id,
    )

    wanderer_user = WandererAccount.objects.get(pk=wanderer_user_id)
    wanderer_managed_map = WandererManagedMap.objects.get(pk=wanderer_managed_map_id)
    logger.debug(wanderer_user, wanderer_managed_map)

    characters_on_acl_ids_set = set(
        wanderer_managed_map.get_character_ids_on_access_list()
    )
    logger.debug(characters_on_acl_ids_set)
    user_character_ids_set = set(wanderer_user.get_all_character_ids())
    logger.debug(user_character_ids_set)

    character_ids_to_remove = characters_on_acl_ids_set & user_character_ids_set

    for character_id_to_remove in character_ids_to_remove:
        logger.debug(
            "Removing char id %d from map %s",
            character_id_to_remove,
            wanderer_managed_map,
        )
        wanderer_managed_map.remove_member_from_access_list(character_id_to_remove)


@shared_task
def cleanup_access_list(wanderer_managed_map_id: int):
    """
    Removes all unwanted members and add missing alts from the map's access list
    """
    logger.info("Updating access list of map id %d", wanderer_managed_map_id)

    wanderer_managed_map = WandererManagedMap.objects.get(pk=wanderer_managed_map_id)

    characters_on_acl_ids_set = set(
        wanderer_managed_map.get_character_ids_on_access_list()
    )
    logger.debug(characters_on_acl_ids_set)
    characters_that_should_be_on_acls_ids_set = set(
        wanderer_managed_map.get_all_accounts_characters_ids()
    )
    logger.debug(characters_that_should_be_on_acls_ids_set)

    character_ids_to_remove = (
        characters_on_acl_ids_set - characters_that_should_be_on_acls_ids_set
    )
    logger.debug(character_ids_to_remove)
    logger.info("Removing %d character ids from the ACL", len(character_ids_to_remove))
    for character_id_to_remove in character_ids_to_remove:
        logger.debug("Removing char id %d", character_id_to_remove)
        wanderer_managed_map.remove_member_from_access_list(character_id_to_remove)

    character_ids_to_add = (
        characters_that_should_be_on_acls_ids_set - characters_on_acl_ids_set
    )
    logger.debug(character_ids_to_add)
    logger.info("Adding %d character ids to the ACL", len(character_ids_to_add))
    for character_id_to_add in character_ids_to_add:
        logger.debug("Adding character id %d", character_id_to_add)
        wanderer_managed_map.add_character_to_acl(character_id_to_add)


@shared_task
def cleanup_all_access_lists():
    """
    Periodically cycle through all access lists to clean up unwanted members and add missing alts
    """
    logger.info("Starting a cleanup of all access lists")

    wanderer_managed_maps = WandererManagedMap.objects.all()

    logger.info("%d maps to cleanup", wanderer_managed_maps.count())

    tasks = [
        cleanup_access_list.si(wanderer_managed_map.id)
        for wanderer_managed_map in wanderer_managed_maps
    ]

    chain(tasks).delay()



"""
Auth hooks with support for dynamic hooks (as opposed to the usual statically defined hooks in the code).

This part still requires a lot of testing

Code ~~stolen~~inspired by
https://github.com/Solar-Helix-Independent-Transport/allianceauth-discord-multiverse/blob/5dff96ddc783ae1a6752e500cab9bd432e653a97/aadiscordmultiverse/auth_hooks.py

"""

from django.db.models.signals import post_delete, post_save
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.notifications.models import Notification
from allianceauth.services.hooks import ServicesHook, get_extension_logger

from .models import WandererManagedMap
from .urls import urlpatterns

logger = get_extension_logger(__name__)


class WandererManagedMapService(ServicesHook):
    """Service for managing a wanderer map ACL with the auth"""

    def __init_subclass__(cls, managed_map: WandererManagedMap):
        super().__init_subclass__()
        cls.managed_map = managed_map

    def __init__(self):
        ServicesHook.__init__(self)
        self.urlpatterns = urlpatterns
        self.access_perm = "wanderer.basic_access"

        self.name = f"wmm:{self.managed_map.wanderer_url}/{self.managed_map.map_slug}"

        self.service_ctrl_template = "wanderer/wmm_services_ctrl.html"

    def render_services_ctrl(self, request):
        return render_to_string(
            self.service_ctrl_template,
            {
                "title": self.managed_map.name,
                "url": str(self.managed_map),
                "has_account": self.managed_map.user_has_account(request.user),
                "map_id": self.managed_map.id,
            },
        )

    def show_service_ctrl(self, user):
        return self.managed_map.accessible_by(user)

    def delete_user(self, user, notify_user=False) -> bool:
        map_name = self.managed_map.name
        logger.info("Deleting user %s from map %s", user, map_name)
        try:
            self.managed_map.delete_user(user)
            if notify_user:
                Notification.objects.notify_user(
                    user,
                    _(f"Account removed from {map_name}"),
                    _(
                        f"Your characters have been removed from the wanderer map {map_name}"
                    ),
                    Notification.Level.WARNING,
                )
            return True
        except Exception as e:
            logger.error("Couldn't delete the user properly: %s", e)
            return False

    def validate_user(self, user):
        logger.debug("Validating user %s account on %s", user, self)
        if self.managed_map.user_has_account(
            user
        ) and not self.managed_map.accessible_by(user):
            logger.info("Removing user %s account on %s", user, self)
            self.delete_user(user, notify_user=True)


def add_del_callback(*args, **kwargs):
    """
    This works great at startup of auth, however has a bug where changes
    made during operation are only captured on a single thread.
    TLDR restart auth after adding a new server
    """
    # Get a list of all map info to check in our hook list
    map_add = list(WandererManagedMap.objects.all())
    # Spit out the ID's for troubleshooting
    logger.info(f"Processing Maps {map_add}")

    # Loop all services and look for our specific hook classes
    for h in hooks._hooks.get("services_hook", []):
        if isinstance(h(), WandererManagedMapService):
            # This is our hook
            # h is an instanced WandererManagedMapService hook with a wanderermap
            if h.managed_map in map_add:
                # this is a known map so remove it from our list of knowns
                map_add.remove(h.managed_map)
            else:
                # This one was deleted remove the hook.
                del h

    # Loop to setup what is missing ( or everything on first boot )
    for map in map_add:
        # What guild_id
        logger.info(f"Adding map {map}")
        # This is the magic to instance the hook class with a new Class Name
        # this way there are no conflicts at runtime
        guild_class = type(
            f"WandererManagedMapService'{map}'",  # New class name
            (WandererManagedMapService,),
            {},  # Super class
            managed_map=map,
        )
        # This adds the hook to the services_hook group to be loaded when needed.
        hooks.register("services_hook", guild_class)


post_save.connect(add_del_callback, sender=WandererManagedMap)
post_delete.connect(add_del_callback, sender=WandererManagedMap)



import responses
from responses import matchers

from django.test import TestCase

from wanderer.wanderer import (
    BadAPIKeyError,
    NotFoundError,
    OwnerEveIdDoesNotExistError,
    add_character_to_acl,
    create_acl_associated_to_map,
    get_acl_members,
    remove_member_from_access_list,
)


class TestApi(TestCase):

    @responses.activate
    def test_create_acl_linked_to_map(self):
        responses.post(
            "http://wanderer.localhost/api/map/acls",
            match=[
                matchers.json_params_matcher(
                    {
                        "acl": {
                            "name": "AA ACL test",
                            "description": "Access list managed by aa-wanderer for the map test. Do not manually edit.",
                            "owner_eve_id": "1",
                        }
                    }
                ),
                matchers.header_matcher({"Authorization": "Bearer dummy-api-key"}),
                matchers.query_param_matcher({"slug": "test"}),
            ],
            json={
                "data": {
                    "id": "095be4ad-6ab4-4024-89b1-7846e25132c6",
                    "name": "test - Auth managed Access list",
                    "description": "Access list managed by alliance auth. Do not manually edit.",
                    "members": [],
                    "inserted_at": "2025-02-17T21:24:44.454134Z",
                    "updated_at": "2025-02-17T21:24:44.454134Z",
                    "api_key": "4620ba44-0930-4410-a513-b4cf51780cfb",
                    "owner_id": "aa49f1b7-35de-4fee-9a9a-8165d52ca8bc",
                }
            },
        )

        api_id, api_key = create_acl_associated_to_map(
            "http://wanderer.localhost", "test", 1, "dummy-api-key"
        )

        self.assertEqual("095be4ad-6ab4-4024-89b1-7846e25132c6", api_id)
        self.assertEqual("4620ba44-0930-4410-a513-b4cf51780cfb", api_key)

    @responses.activate
    def test_create_acl_wrong_token(self):
        responses.post(
            "http://wanderer.localhost/api/map/acls",
            match=[
                matchers.json_params_matcher(
                    {
                        "acl": {
                            "name": "AA ACL test",
                            "description": "Access list managed by aa-wanderer for the map test. Do not manually edit.",
                            "owner_eve_id": "1",
                        }
                    }
                ),
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
                matchers.query_param_matcher({"slug": "test"}),
            ],
            status=401,
        )

        self.assertRaises(
            BadAPIKeyError,
            create_acl_associated_to_map,
            "http://wanderer.localhost",
            "test",
            1,
            "bad-api-key",
        )

    @responses.activate
    def test_get_character_ids_on_access_list(self):
        responses.get(
            "http://wanderer.localhost/api/acls/ACL_UUID",
            match=[
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
            ],
            json={
                "data": {
                    "id": "ACL_UUID",
                    "name": "AA ACL test",
                    "description": "Access list managed by aa-wanderer for the map test. Do not manually edit.",
                    "members": [
                        {
                            "id": "072d2059-5e30-4323-ab5b-2ee3d0bf3466",
                            "name": "Feh'dow Rokym",
                            "inserted_at": "2025-03-03T22:09:15.242731Z",
                            "updated_at": "2025-03-03T22:09:15.242731Z",
                            "role": "viewer",
                            "eve_character_id": "2116864032",
                        },
                        {
                            "id": "b6dba477-b74d-48ee-9de8-caf3dd159605",
                            "name": "Tyd Drakken",
                            "inserted_at": "2025-03-03T22:23:14.736920Z",
                            "updated_at": "2025-03-03T22:23:14.736920Z",
                            "role": "viewer",
                            "eve_character_id": "626646627",
                        },
                        {
                            "id": "e72ea6d7-65cc-4eae-85b5-c43d060084e8",
                            "name": "T'rahk Rokym",
                            "inserted_at": "2025-03-03T22:35:31.149913Z",
                            "updated_at": "2025-03-03T22:35:31.149913Z",
                            "role": "viewer",
                            "eve_character_id": "2112073677",
                        },
                    ],
                    "inserted_at": "2025-03-03T21:07:29.784792Z",
                    "updated_at": "2025-03-03T21:07:29.784792Z",
                    "api_key": "bad-api-key",
                    "owner_id": "aa49f1b7-35de-4fee-9a9a-8165d52ca8bc",
                }
            },
        )

        characters_on_acl = get_acl_members(
            "http://wanderer.localhost", "ACL_UUID", "bad-api-key"
        )

        self.assertIn(2116864032, characters_on_acl)
        self.assertIn(626646627, characters_on_acl)
        self.assertIn(2112073677, characters_on_acl)

    @responses.activate
    def test_get_characters_ids_when_corporation(self):
        """
        Corporation will return a null value in the eve_character_id field
        TODO this null value is a bug and should be fixed at some point
        """
        responses.get(
            "http://wanderer.localhost/api/acls/ACL_UUID",
            match=[
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
            ],
            json={
                "data": {
                    "id": "ACL_UUID",
                    "name": "AA ACL test",
                    "description": "Access list managed by aa-wanderer for the map test. Do not manually edit.",
                    "members": [
                        {
                            "id": "072d2059-5e30-4323-ab5b-2ee3d0bf3466",
                            "name": "Feh'dow Rokym",
                            "inserted_at": "2025-03-03T22:09:15.242731Z",
                            "updated_at": "2025-03-03T22:09:15.242731Z",
                            "role": "viewer",
                            "eve_character_id": "2116864032",
                        },
                        {
                            "id": "b6dba477-b74d-48ee-9de8-caf3dd159605",
                            "name": "Tyd Drakken",
                            "inserted_at": "2025-03-03T22:23:14.736920Z",
                            "updated_at": "2025-03-03T22:23:14.736920Z",
                            "role": "viewer",
                            "eve_character_id": "626646627",
                        },
                        {
                            "id": "e72ea6d7-65cc-4eae-85b5-c43d060084e8",
                            "name": "T'rahk Rokym",
                            "inserted_at": "2025-03-03T22:35:31.149913Z",
                            "updated_at": "2025-03-03T22:35:31.149913Z",
                            "role": "viewer",
                            "eve_character_id": "2112073677",
                        },
                        {
                            "id": "f96069d3-daad-423f-8b9e-f3973871365f",
                            "name": "Rokym's managment organisation",
                            "inserted_at": "2025-03-03T22:47:27.404057Z",
                            "updated_at": "2025-03-03T22:47:27.404057Z",
                            "role": "viewer",
                            "eve_character_id": None,
                        },
                    ],
                    "inserted_at": "2025-03-03T21:07:29.784792Z",
                    "updated_at": "2025-03-03T21:07:29.784792Z",
                    "api_key": "bad-api-key",
                    "owner_id": "aa49f1b7-35de-4fee-9a9a-8165d52ca8bc",
                }
            },
        )

        characters_on_acl = get_acl_members(
            "http://wanderer.localhost", "ACL_UUID", "bad-api-key"
        )

        self.assertIn(2116864032, characters_on_acl)
        self.assertIn(626646627, characters_on_acl)
        self.assertIn(2112073677, characters_on_acl)

    @responses.activate
    def test_add_character_to_access_list(self):
        responses.post(
            "http://wanderer.localhost/api/acls/ACL_UUID/members",
            match=[
                matchers.json_params_matcher(
                    {
                        "member": {
                            "eve_character_id": "2112073677",
                            "role": "viewer",
                        }
                    }
                ),
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
            ],
            json={
                "data": {
                    "id": "4b143d10-1232-4ac8-a909-46276a4a6064",
                    "name": "T'rahk Rokym",
                    "inserted_at": "2025-03-03T22:29:53.334988Z",
                    "updated_at": "2025-03-03T22:29:53.334988Z",
                    "role": "viewer",
                    "eve_character_id": "2112073677",
                }
            },
        )

        add_character_to_acl(
            "http://wanderer.localhost", "ACL_UUID", "bad-api-key", 2112073677
        )

    @responses.activate
    def test_remove_member_from_access_list(self):
        responses.delete(
            "http://wanderer.localhost/api/acls/ACL_UUID/members/1000",
            match=[
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
            ],
            json={
                "ok": "true",
            },
        )

        remove_member_from_access_list(
            "http://wanderer.localhost", "ACL_UUID", "bad-api-key", 1000
        )

    @responses.activate
    def test_remove_member_not_on_access_list(self):
        responses.delete(
            "http://wanderer.localhost/api/acls/ACL_UUID/members/1000",
            match=[
                matchers.header_matcher({"Authorization": "Bearer bad-api-key"}),
            ],
            status=404,
            json={"error": "Membership not found for given ACL and external id"},
        )

        self.assertRaises(
            NotFoundError,
            remove_member_from_access_list,
            "http://wanderer.localhost",
            "ACL_UUID",
            "bad-api-key",
            1000,
        )

    @responses.activate
    def test_unknown_owner_eve_id(self):
        """Error will be returned if an unknown owner_eve_id is given when creating an access list"""
        responses.post(
            "http://wanderer.localhost/api/map/acls?slug=map_slug",
            match=[matchers.header_matcher({"Authorization": "Bearer bad-api-key"})],
            status=400,
            json={
                "error": '{:error, "owner_eve_id does not match any existing character"}'
            },
        )

        self.assertRaises(
            OwnerEveIdDoesNotExistError,
            create_acl_associated_to_map,
            "http://wanderer.localhost",
            "map_slug",
            1001,
            "bad-api-key",
        )



"""Interactions with the wanderer's API"""

import requests

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class BadAPIKeyError(Exception):
    """Exception raised when a wrong API key is provided"""


class NotFoundError(Exception):
    """Exception raised when the API returned an expected 404"""


class OwnerEveIdDoesNotExistError(Exception):
    """Exception raised when attempting to create a map with an owner not known by Wanderer"""


DEFAULT_TIMEOUT = 5


def create_acl_associated_to_map(
    wanderer_url: str, map_slug: str, requesting_character_id: int, map_api_key: str
) -> (str, str):
    """
    Will create a new ACL associated with the map `map_slug`

    Returns the ACL associated id and API key
    """

    logger.info(
        "Creating ACL on wanderer %s for map %s by character %d with api key %s",
        wanderer_url,
        map_slug,
        requesting_character_id,
        map_api_key,
    )

    r = requests.post(
        f"{wanderer_url}/api/map/acls?slug={map_slug}",
        headers={"Authorization": f"Bearer {map_api_key}"},
        json={
            "acl": {
                "name": f"AA ACL {map_slug}",
                "description": f"Access list managed by aa-wanderer for the map {map_slug}. Do not manually edit.",
                "owner_eve_id": str(requesting_character_id),
            }
        },
        timeout=DEFAULT_TIMEOUT,
    )

    logger.debug("Received status code %d", r.status_code)
    logger.debug(r.text)

    if (
        r.status_code == 400
        and "owner_eve_id does not match any existing character" in r.text
    ):
        raise OwnerEveIdDoesNotExistError(
            f"The eve character with id {requesting_character_id} "
            "doesn't seem to be known by Wanderer"
        )

    if r.status_code == 401:
        raise BadAPIKeyError(
            f"The API key {map_api_key} returned a 401 when trying to create an ACL on map {wanderer_url} {map_slug}"
        )

    r.raise_for_status()

    acl_id = r.json()["data"]["id"]
    acl_key = r.json()["data"]["api_key"]
    logger.info("Successfully created ACL id %s")

    return acl_id, acl_key


def get_acl_members(wanderer_url: str, acl_id: str, acl_api_key: str) -> list[int]:
    """
    Returns all members eve_character_id present in an ACL
    """
    logger.info("Requesting character on the ACL of map %s / %s", wanderer_url, acl_id)

    r = requests.get(
        f"{wanderer_url}/api/acls/{acl_id}",
        headers={"Authorization": f"Bearer {acl_api_key}"},
        timeout=DEFAULT_TIMEOUT,
    )
    logger.debug(r)
    logger.debug(r.text)

    if r.status_code == 401:
        raise BadAPIKeyError(
            f"The API key {acl_api_key} returned a 401 when trying to access the members of ACL {wanderer_url} {acl_id}"
        )

    r.raise_for_status()

    return [
        int(member["eve_character_id"])
        for member in r.json()["data"]["members"]
        if member["eve_character_id"]
    ]


def add_character_to_acl(
    wanderer_url: str, acl_id: str, acl_api_key: str, character_id: int
):
    """
    Adds a single character to the ACL with the viewer role
    """

    r = requests.post(
        f"{wanderer_url}/api/acls/{acl_id}/members",
        headers={"Authorization": f"Bearer {acl_api_key}"},
        json={
            "member": {
                "eve_character_id": str(character_id),
                "role": "viewer",
            }
        },
        timeout=DEFAULT_TIMEOUT,
    )

    if r.status_code == 401:
        raise BadAPIKeyError(
            f"The API key {acl_api_key} returned a 401 when trying to access the members of ACL {wanderer_url} {acl_id}"
        )

    r.raise_for_status()


def remove_member_from_access_list(
    wanderer_url: str, acl_id: str, acl_api_key: str, member_id: int
):
    """
    Removes the member with specified id from the ACL
    """

    r = requests.delete(
        f"{wanderer_url}/api/acls/{acl_id}/members/{member_id}",
        headers={"Authorization": f"Bearer {acl_api_key}"},
        timeout=DEFAULT_TIMEOUT,
    )

    if r.status_code == 401:
        raise BadAPIKeyError(
            f"The API key {acl_api_key} returned a 401 when trying to access the members of ACL {wanderer_url} {acl_id}"
        )

    if r.status_code == 404:  # If the API isn't found a 401 is raised
        raise NotFoundError(f"Member id {member_id} was not found on ACL {acl_id}")

    r.raise_for_status()
